# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Defines the implementation of the base Computation interface."""

from typing import Optional

from federated_language.compiler import building_blocks
from federated_language.computation import computation_base
from federated_language.computation import function_utils
from federated_language.context_stack import context_stack_impl
from federated_language.proto import computation_pb2
from federated_language.types import computation_types


class ConcreteComputation(computation_base.Computation):
  """A representation of a `pb.Computation` in the `federated_language.Computation` interface.

  This implementation exposes methods to retrieve the backing `pb.Computation`,
  as well as the Python representation of this protocol buffer represented by
  an instance of `building_blocks.ComputationBuildingBlock`.
  """

  @classmethod
  def from_building_block(
      cls, building_block: building_blocks.ComputationBuildingBlock
  ) -> 'ConcreteComputation':
    """Converts a computation building block to a computation impl."""
    return cls(
        computation_proto=building_block.to_proto(),
        context_stack=context_stack_impl.context_stack,
        annotated_type=building_block.type_signature,  # pytype: disable=wrong-arg-types
    )

  def to_building_block(self):
    # TODO: b/161560999 - currently destroys annotated type.
    # This should perhaps be fixed by adding `type_parameter` to `from_proto`.
    return building_blocks.ComputationBuildingBlock.from_proto(self._proto)

  def to_compiled_building_block(self):
    return building_blocks.CompiledComputation(
        self._proto, type_signature=self.type_signature
    )

  def __init__(
      self,
      *,
      computation_proto: computation_pb2.Computation,
      context_stack: context_stack_impl.ContextStack,
      annotated_type: Optional[computation_types.FunctionType] = None,
  ):
    """Constructs a new instance of ConcreteComputation from the computation_proto.

    Args:
      computation_proto: The protocol buffer that represents the computation, an
        instance of computation_pb2.Computation.
      context_stack: The context stack to use.
      annotated_type: Optional, type information with additional annotations
        that replaces the information in `computation_proto.type`.

    Raises:
      TypeError: If `annotated_type` is not `None` and is not compatible with
        `computation_proto.type`.
      ValueError: If `computation_proto.type` is `None`.
    """
    if computation_proto.type is None:
      raise ValueError('Expected `computation_proto.type` to not be `None`.')
    type_spec = computation_types.Type.from_proto(computation_proto.type)

    if annotated_type is not None:
      if type_spec is None or not type_spec.is_assignable_from(annotated_type):
        raise TypeError(
            'annotated_type not compatible with computation_proto.type\n'
            f'computation_proto.type: {type_spec}\n'
            f'annotated_type: {annotated_type}'
        )
      type_spec = annotated_type

    if not isinstance(type_spec, computation_types.FunctionType):
      raise TypeError(
          f'{type_spec} is not a functional type, from proto: '
          f'{computation_proto}'
      )

    self._type_signature = type_spec
    self._context_stack = context_stack
    self._proto = computation_proto

  @classmethod
  def from_proto(
      cls, computation_pb: computation_pb2.Computation
  ) -> 'ConcreteComputation':
    """Returns a `ConcreteComputation` for the `computation_pb`."""
    return ConcreteComputation(
        computation_proto=computation_pb,
        context_stack=context_stack_impl.context_stack,
    )

  def to_proto(self) -> computation_pb2.Computation:
    """Returns a `computation_pb2.Computation` for this computation."""
    return self._proto

  @property
  def type_signature(self) -> computation_types.FunctionType:
    return self._type_signature

  @property
  def context_stack(self) -> context_stack_impl.ContextStack:
    return self._context_stack

  def __call__(self, *args, **kwargs):
    arg = function_utils.pack_args(self._type_signature.parameter, args, kwargs)
    result = self._context_stack.current.invoke(self, arg)
    return result

  def __eq__(self, other: object) -> bool:
    if self is other:
      return True
    elif not isinstance(other, ConcreteComputation):
      return NotImplemented
    return self._proto == other._proto

  def __hash__(self) -> int:
    return hash(self._proto.SerializeToString(deterministic=True))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A library of static analysis functions for building blocks."""

from federated_language.compiler import building_blocks
from federated_language.compiler import intrinsic_defs


def is_called_intrinsic(comp, uri=None):
  """Tests if `comp` is a called intrinsic with the given `uri`.

  Args:
    comp: The computation building block to test.
    uri: An optional URI or list of URIs; the same form as what is accepted by
      isinstance.

  Returns:
    `True` if `comp` is a called intrinsic with the given `uri`, otherwise
    `False`.
  """
  if isinstance(uri, str):
    uri = [uri]
  return (
      isinstance(comp, building_blocks.Call)
      and isinstance(comp.function, building_blocks.Intrinsic)
      and (uri is None or comp.function.uri in uri)
  )


def is_called_aggregation(
    building_block: building_blocks.ComputationBuildingBlock,
    kind: intrinsic_defs.AggregationKind,
):
  return (
      isinstance(building_block, building_blocks.Call)
      and isinstance(building_block.function, building_blocks.Intrinsic)
      and building_block.function.intrinsic_def().aggregation_kind is kind
  )


def is_identity_function(comp):
  """Returns `True` if `comp` is an identity function, otherwise `False`."""
  return (
      isinstance(comp, building_blocks.Lambda)
      and isinstance(comp.result, building_blocks.Reference)
      and comp.parameter_name == comp.result.name
  )



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The Federated Language library."""

# pylint: disable=g-importing-member
from federated_language import framework
from federated_language import program
from federated_language.common_libs.serializable import Serializable
from federated_language.compiler.array import Array
from federated_language.compiler.array import from_proto as array_from_proto
from federated_language.compiler.array import from_proto_content as array_from_proto_content
from federated_language.compiler.array import is_compatible_dtype as array_is_compatible_dtype
from federated_language.compiler.array import is_compatible_shape as array_is_compatible_shape
from federated_language.compiler.array import to_proto as array_to_proto
from federated_language.compiler.array import to_proto_content as array_to_proto_content
from federated_language.computation.computation_base import Computation
from federated_language.federated_context.federated_computation import federated_computation
from federated_language.federated_context.intrinsics import federated_aggregate
from federated_language.federated_context.intrinsics import federated_broadcast
from federated_language.federated_context.intrinsics import federated_eval
from federated_language.federated_context.intrinsics import federated_map
from federated_language.federated_context.intrinsics import federated_max
from federated_language.federated_context.intrinsics import federated_mean
from federated_language.federated_context.intrinsics import federated_min
from federated_language.federated_context.intrinsics import federated_secure_select
from federated_language.federated_context.intrinsics import federated_secure_sum
from federated_language.federated_context.intrinsics import federated_secure_sum_bitwidth
from federated_language.federated_context.intrinsics import federated_select
from federated_language.federated_context.intrinsics import federated_sum
from federated_language.federated_context.intrinsics import federated_value
from federated_language.federated_context.intrinsics import federated_zip
from federated_language.federated_context.intrinsics import sequence_map
from federated_language.federated_context.intrinsics import sequence_reduce
from federated_language.federated_context.intrinsics import sequence_sum
from federated_language.federated_context.value_impl import to_value
from federated_language.federated_context.value_impl import Value
from federated_language.types.array_shape import ArrayShape
from federated_language.types.array_shape import from_proto as array_shape_from_proto
from federated_language.types.array_shape import is_compatible_with as array_shape_is_compatible_with
from federated_language.types.array_shape import is_shape_fully_defined as array_shape_is_fully_defined
from federated_language.types.array_shape import is_shape_scalar as array_shape_is_scalar
from federated_language.types.array_shape import num_elements_in_shape as num_elements_in_array_shape
from federated_language.types.array_shape import to_proto as array_shape_to_proto
from federated_language.types.computation_types import AbstractType
from federated_language.types.computation_types import FederatedType
from federated_language.types.computation_types import FunctionType
from federated_language.types.computation_types import PlacementType
from federated_language.types.computation_types import SequenceType
from federated_language.types.computation_types import StructType
from federated_language.types.computation_types import StructWithPythonType
from federated_language.types.computation_types import TensorType
from federated_language.types.computation_types import to_type
from federated_language.types.computation_types import Type
from federated_language.types.dtype_utils import can_cast as can_cast_dtype
from federated_language.types.dtype_utils import from_proto as dtype_from_proto
from federated_language.types.dtype_utils import infer_dtype
from federated_language.types.dtype_utils import is_valid_dtype
from federated_language.types.dtype_utils import to_proto as dtype_to_proto
from federated_language.types.placements import CLIENTS
from federated_language.types.placements import SERVER
from federated_language.types.typed_object import TypedObject
from federated_language.version import __version__
# pylint: enable=g-importing-member



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Defines abstract interfaces for representing references to values.

These abstract interfaces provide the capability to handle values without
requiring them to be materialized as Python objects. Instances of these
abstract interfaces represent values of type `federated_language.TensorType` and
can be placed
on the server, elements of structures that are placed on the server, or
unplaced.
"""

import abc
import asyncio
from collections.abc import Iterable
from typing import Union

from federated_language.program import structure_utils
from federated_language.types import computation_types
from federated_language.types import typed_object
import numpy as np

MaterializableTypeSignature = Union[
    computation_types.TensorType,
    computation_types.SequenceType,
]
_MaterializedArrayValue = Union[
    # Python types
    bool,
    int,
    float,
    complex,
    str,
    bytes,
    # Numpy types
    np.generic,
    np.ndarray,
]
MaterializedValue = Union[
    _MaterializedArrayValue,
    Iterable[_MaterializedArrayValue],
]
MaterializedStructure = structure_utils.Structure[MaterializedValue]
MaterializableValue = Union[
    MaterializedValue,
    'MaterializableValueReference',
]
MaterializableStructure = structure_utils.Structure[MaterializableValue]


class MaterializableValueReference(typed_object.TypedObject, abc.ABC):
  """An abstract interface representing references to server-placed values."""

  @property
  @abc.abstractmethod
  def type_signature(self) -> MaterializableTypeSignature:
    """The `federated_language.Type` of this object."""
    raise NotImplementedError

  @abc.abstractmethod
  async def get_value(self) -> MaterializedValue:
    """Returns the referenced value.

    The Python type of the referenced value depends on the `type_signature`:

    | Federated Language Type  | Python Type                                  |
    | ------------------------ | -------------------------------------------- |
    | `TensorType`             | `bool`, `int`, `float`, `complex`, `str`,    |
    |                          | `bytes`, `np.generic`, or `np.ndarray`       |
    | `SequenceType`           | `Iterable` of any Python type corresponding  |
    |                          |  to a `federated_language.TensorType`        |
    """
    raise NotImplementedError


async def materialize_value(
    value: MaterializableStructure,
) -> MaterializedStructure:
  """Materializes the `federated_language.program.MaterializableValueReference`s in `value`.

  Args:
    value: A `federated_language.program.MaterializableStructure` to
      materialize.

  Returns:
    A `federated_language.program.MaterializedStructure`.
  """

  async def _materialize(value: MaterializableValue) -> MaterializedValue:
    if isinstance(value, MaterializableValueReference):
      return await value.get_value()
    else:
      return value

  flattened_value = structure_utils.flatten(value)
  materialized_value = await asyncio.gather(
      *[_materialize(v) for v in flattened_value]
  )
  return structure_utils.unflatten_as(value, materialized_value)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for releasing values from a federated program to logs."""

from typing import Optional

from absl import logging
from federated_language.program import release_manager
from federated_language.program import value_reference


class LoggingReleaseManager(
    release_manager.ReleaseManager[
        release_manager.ReleasableStructure, release_manager.Key
    ]
):
  """A `federated_language.program.ReleaseManager` that releases values to logs.

  A `federated_language.program.LoggingReleaseManager` is a utility for
  releasing values from a
  federated program to logs and is used to release values from platform storage
  to customer storage in a federated program.

  Values are released to logs as string representations of Python objects. When
  the value is released, if the value is a value reference or a structure
  containing value references, each value reference is materialized.
  """

  async def release(
      self,
      value: release_manager.ReleasableStructure,
      key: Optional[release_manager.Key],
  ) -> None:
    """Releases `value` from a federated program.

    Args:
      value: A `federated_language.program.ReleasableStructure` to release.
      key: An optional value used to reference the released `value`.
    """
    materialized_value = await value_reference.materialize_value(value)
    logging.info('Releasing')
    logging.info('  value: %s', materialized_value)
    if key is not None:
      logging.info('  key: %s', key)



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility functions for checking Python types."""

from collections.abc import Sequence
import sys
import typing
from typing import Optional, Protocol, TypeVar, Union

from typing_extensions import TypeGuard


@typing.runtime_checkable
class SupportsNamedTuple(Protocol):
  """A `typing.Protocol` with two abstract method `_fields` and `_asdict`."""

  @property
  def _fields(self) -> tuple[str, ...]:
    ...

  def _asdict(self) -> dict[str, object]:
    ...


_NT = TypeVar('_NT', bound=Optional[str])
_VT = TypeVar('_VT', bound=object)


def is_name_value_pair(
    obj: object,
    name_type: type[_NT] = Optional[str],
    value_type: type[_VT] = object,
) -> TypeGuard[tuple[_NT, _VT]]:
  """Returns `True` if `obj` is a name value pair, otherwise `False`.

  In `federated_language`, a name value pair (or named field) is a
  `collection.abc.Sequence` of exactly two elements, a `name` (which can be
  `None`) and a `value`.

  Args:
    obj: The object to test.
    name_type: The type of the name.
    value_type: The type of the value.
  """
  if not isinstance(obj, Sequence) or len(obj) != 2:
    return False
  name, value = obj

  # Before Python 3.10, you could not pass a `Union Type` to isinstance, see
  # https://docs.python.org/3/library/functions.html#isinstance.
  if sys.version_info < (3, 10):

    def _unpack_type(x):
      origin = typing.get_origin(x)
      if origin is Union:
        return typing.get_args(name_type)
      elif origin is not None:
        return origin
      else:
        return x

    name_type = _unpack_type(name_type)
    value_type = _unpack_type(value_type)

  return isinstance(name, name_type) and isinstance(value, value_type)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for saving and loading program state in a federated program."""

import abc
from typing import Generic, Optional, TypeVar, Union

from federated_language.common_libs import serializable
from federated_language.program import structure_utils
from federated_language.program import value_reference

# ProgramStateManager's may save any value in addition to materializable values.
ProgramStateValue = Union[
    object,
    value_reference.MaterializableValue,
    # `federated_language.Serializable` objects are intended to be impemented by
    # platform authors and used by program logic authors; it should not be
    # common for program logic authors to implement
    # `federated_language.Serializable` objects.
    serializable.Serializable,
]
ProgramStateStructure = TypeVar(
    'ProgramStateStructure',
    bound=structure_utils.Structure[ProgramStateValue],
)


class ProgramStateExistsError(Exception):
  """Raised when the program state already exists."""

  def __init__(self, *, version: int, path: str):
    super().__init__(
        f'Program state already exists for version [{version}] at path'
        f' [{path}].'
    )


class ProgramStateNotFoundError(Exception):
  """Raised when the program state cannot be found."""

  def __init__(self, version: int):
    super().__init__(f'No program state found for version: {version}.')


class ProgramStateManager(abc.ABC, Generic[ProgramStateStructure]):
  """An interface for saving and loading program state in a federated program.

  A `federated_language.program.ProgramStateManager` is used to implement fault
  tolerance in a
  federated program. The structure or type of the program state that is saved is
  unknown at construction time and can change as the program runs.
  """

  @abc.abstractmethod
  async def get_versions(self) -> Optional[list[int]]:
    """Returns a list of saved versions or `None`.

    Returns:
      A list of saved versions or `None` if there is no saved program state.
    """
    raise NotImplementedError

  @abc.abstractmethod
  async def load(
      self, version: int, structure: ProgramStateStructure
  ) -> ProgramStateStructure:
    """Returns the saved program state for the given `version`.

    Args:
      version: A integer representing the version of a saved program state.
      structure: The structure of the saved program state for the given
        `version` used to support serialization and deserailization of
        user-defined classes in the structure.

    Raises:
      ProgramStateNotFoundError: If there is no program state for the given
        `version`.
    """
    raise NotImplementedError

  async def load_latest(
      self, structure: ProgramStateStructure
  ) -> tuple[ProgramStateStructure, int]:
    """Returns the latest saved program state and version or (`None`, 0).

    Args:
      structure: The structure of the saved program state for the given
        `version` used to support serialization and deserailization of
        user-defined classes in the structure.

    Returns:
      A tuple of the latest saved (program state, version) or (`None`, 0) if
      there is no latest saved program state.
    """
    versions = await self.get_versions()
    if versions is None or not versions:
      return None, 0
    latest_version = max(versions)
    try:
      return await self.load(latest_version, structure), latest_version
    except ProgramStateNotFoundError:
      return None, 0

  @abc.abstractmethod
  async def save(
      self, program_state: ProgramStateStructure, version: int
  ) -> None:
    """Saves `program_state` for the given `version`.

    Args:
      program_state: A `federated_language.program.ProgramStateStructure` to
        save.
      version: A strictly increasing integer representing the version of a saved
        `program_state`.

    Raises:
      ProgramStateExistsError: If there is already program state for the given
        `version`.
    """
    raise NotImplementedError
